//this file is part of eMule
//Copyright (C)2002 Merkur ( devs@emule-project.net / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

// BEGIN MOD netfinity:
#include "UploadQueue.h"
#include "EMSocket.h"
#include "ClientUDPSocket.h"
#include "Exceptions.h"
// END MOD netfinity

#include "NetF.h"

void CDownloadQueue::Process(){
	try{
	ProcessLocalRequests(); // send src requests to local server
	}
	CATCH_DFLT_EXCEPTIONS(_T("CDownloadQueue::Process (ProcessLocalRequests)"))

	uint32 downspeed = 0;
    CBytesPerSecond maxDownload = thePrefs.GetMaxDownloadInBytesPerSec(true);
	static CTick tickLastLoop;
	static CTick tickLastLog;
	CTick tickThisLoop = CTick::GetCurrentTick();

	// MOD BEGIN netfinity: Session limit, are we leeching?
	// Don't download more than 3 times our upload (excl. friend upload)
	float	sessionUlDlRatio = ((float) (theStats.sessionSentBytes - theStats.sessionSentBytesToFriend) + 1.0) / ((float) theStats.sessionReceivedBytes + 1.0);
	if(sessionUlDlRatio < 0.40) // Activate throttling if UL is less than 40% of DL
	{
		CBytesPerSecond	minDownload = CBytesPerSecond(10240);
		// We are uploading so maybe 10kB/s is to high?
		if(theApp.uploadqueue->GetUploadQueueLength() > 0 && theStats.sessionReceivedBytes >= PARTSIZE)
		{
			minDownload = CBytesPerSecond(10240) / (theStats.sessionReceivedBytes / PARTSIZE);
		}
		CBytesPerSecond	currUpload = min(theApp.uploadqueue->GetToNetworkDatarate(), CBytesPerSecond(4096)); // Exclude overhead and friend upload, but not less than 4 kB/s (we don't want to kill downloads completly)
		maxDownload = min(max(minDownload, (27 * currUpload * sessionUlDlRatio * sessionUlDlRatio) + CBytesPerSecond(3072)), maxDownload);
		if(maxDownload < CBytesPerSecond(1024)) maxDownload = CBytesPerSecond(1024);


		if (thePrefs.GetVerbose() && (!tickLastLog.IsTickSet() || tickLastLog.Elapsed(tickThisLoop) >= MIN2MS(1)))
		{
			tickLastLog = CTick::GetCurrentTick();
			AddDebugLogLine(false, _T("Download rate has been limited to %d kB/s, due to session ratio"), maxDownload / CBytesPerSecond(1024));
		}
		theNetF.m_bDownloadLimited = true;
	}
	else
		theNetF.m_bDownloadLimited = false;
	// MOD END netfinity

	if (maxDownload != CBytesPerSecond::Infinite() && datarate > 1500){
		downspeed = (maxDownload*100)/(datarate+1);
		if (downspeed < 50)
			downspeed = 50;
		else if (downspeed > 200)
			downspeed = 200;
	}

	/*while(avarage_dr_list.GetCount() > 0 && (avarage_dr_list.GetHead().timestamp.Elapsed() > SEC2MS(10)) )
		m_datarateMS-=avarage_dr_list.RemoveHead().datalen.GetValue();
	
	if (avarage_dr_list.GetCount()>1){
		datarate = m_datarateMS / avarage_dr_list.GetCount();
	} else {
		datarate = 0;
	}*/

	CBytesPerSecond datarateX=0;
	udcounter++;

	// MOD BEGIN netfinity: Calculate allowed average download rate
	// Count number of clients in download state
	uint32	downClients = 0;
	uint32	downFiles = 0;
	CBytesPerSecond	allowedDownload = 0;
	uint32	downClientsAvg = 1;
	for (POSITION pos = filelist.GetHeadPosition();pos != 0;)
	{
		uint32	tempClients = 0;
		CPartFile* cur_file = filelist.GetNext(pos);
		tempClients = cur_file->GetTransferringSrcCount();
		downClients += tempClients;
		if(tempClients > 0) ++downFiles;
	}
	if(downClients == 0) downClients = 1;
	if(downspeed == 0)
	{	// UNLIMITED download
		//downspeed = 200;
		allowedDownload = CBytesPerSecond(3072);
		//if(allowedDownload < 3072) allowedDownload = 3072;
		//AddDebugLogLine(false, _T("UNLIMITED datarate = %d, allowedDownload = %d"), datarate, allowedDownload);
	}
	else
	{	// Limited download
		allowedDownload = maxDownload / downClients;
		//if(datarate < allowedDownload + maxDownload)
		//	allowedDownload = allowedDownload + maxDownload - datarate;
		//else
		//	allowedDownload = CBytesPerSecond(200);
		//AddDebugLogLine(false, _T("downspeed = %d, datarate = %d, allowedDownload = %d"), downspeed, datarate, allowedDownload);
	}
	if(allowedDownload < CBytesPerSecond(200)) allowedDownload = CBytesPerSecond(200);
	if(allowedDownload > maxDownload) allowedDownload = maxDownload;
	// MOD END netfinity

	//filelist is already sorted by prio, therefore I removed all the extra loops..
	for (POSITION pos = filelist.GetHeadPosition();pos != 0;){
		CPartFile* cur_file = filelist.GetNext(pos);
		if (cur_file->GetStatus() == PS_READY || cur_file->GetStatus() == PS_EMPTY){
			datarateX += cur_file->Process(downspeed,udcounter,allowedDownload.GetValue());
		}
		else{
			//This will make sure we don't keep old sources to paused and stoped files..
			cur_file->StopPausedFile();
		}
	}

	// MOD BEGIN netfinity: Process received packet queue
	//  This code needs some major remodeling
	ReceivedPacket rcvpacket;
	while (true)
	{
		receivequeueLocker.Lock();
		if (!receivedpacket_queue.IsEmpty())
			rcvpacket = receivedpacket_queue.RemoveHead();
		else
			rcvpacket.socket = NULL; // Don't block longer than necessary
		if (rcvpacket.socket)
		{
			if (rcvpacket.udpPacket)
				reinterpret_cast<CClientUDPSocket*>(rcvpacket.socket)->PacketReceived(rcvpacket.packet, rcvpacket.ip, rcvpacket.port);
			else if (reinterpret_cast<CEMSocket*>(rcvpacket.socket)->IsConnected())
				reinterpret_cast<CEMSocket*>(rcvpacket.socket)->PacketReceived(rcvpacket.packet);
			delete rcvpacket.packet;
			receivequeueLocker.Unlock();
		}
		else
		{
			receivequeueLocker.Unlock();
			break;
		}
	}
	// MOD END netfinity:

	// netfinity: Datarate calculation
	/*if (tickLastLoop.IsTickSet())
		datarate.Update(datarateX * tickLastLoop.Elapsed(tickThisLoop));
	tickLastLoop = tickThisLoop; */
	datarate = datarateX;

	/*TransferredData newitem = {datarateX, CTick::GetCurrentTick()};
	avarage_dr_list.AddTail(newitem);
	m_datarateMS+=datarateX;*/

	if (udcounter == 5){
		if (theApp.serverconnect->IsUDPSocketAvailable()){
		    if(!lastudpstattime.IsTickSet() || lastudpstattime.Elapsed() > UDPSERVERSTATTIME)
			{
			    lastudpstattime.SetCurrentTick();
				try{
			    theApp.serverlist->ServerStats();
				}
				CATCH_DFLT_EXCEPTIONS(_T("CDownloadQueue::Process (theApp.serverlist->ServerStats)"))
		    }
	    }
	}

	if (udcounter == 10){
		udcounter = 0;
		if (theApp.serverconnect->IsUDPSocketAvailable()){
			if (!lastudpsearchtime.IsTickSet() || lastudpsearchtime.Elapsed() > UDPSERVERREASKTIME)
			{
				try{
				SendNextUDPPacket();
				}
				CATCH_DFLT_EXCEPTIONS(_T("CDownloadQueue::Process (SendNextUDPPacket)"))
			}
		}
	}

	CheckDiskspaceTimed();

// ZZ:DownloadManager -->
    //if((!m_dwLastA4AFtime) || (::GetTickCount() - m_dwLastA4AFtime) > 2*60*1000) {
    //    theApp.clientlist->ProcessA4AFClients();
    //    m_dwLastA4AFtime = ::GetTickCount();
    //}
// <-- ZZ:DownloadManager
}
