/** @file		SafePointers.h
 *  @brief	Ensure pointers are safe to use
 *  @author	netfinity
 */
#pragma once

/**
 *  @brief	Ensure pointer is safe to use
 */
template <class _T>
class CSafePtr
{
public:
	CSafePtr()
	{
		m_pTarget = NULL;
	}
	CSafePtr(_T* pTarget)
	{
		m_pTarget = pTarget;
		if (m_pTarget)
			m_pTarget->AddSafePtr(reinterpret_cast<CSafePtr<void>*>(this));
	}
	CSafePtr(CSafePtr<_T>& refSafePtr)
	{
		m_pTarget = refSafePtr->m_pTarget;
		if (m_pTarget)
			m_pTarget->AddSafePtr(reinterpret_cast<CSafePtr<void>*>(this));
	}
	~CSafePtr()
	{
		if (m_pTarget)
			m_pTarget->RemoveSafePtr(reinterpret_cast<CSafePtr<void>*>(this));
	}
	CSafePtr& operator = (_T* pTarget)
	{
		if (m_pTarget)
			m_pTarget->RemoveSafePtr(reinterpret_cast<CSafePtr<void>*>(this));
		m_pTarget = pTarget;
		if (m_pTarget)
			m_pTarget->AddSafePtr(reinterpret_cast<CSafePtr<void>*>(this));
		return *this;
	}
	CSafePtr& operator = (CSafePtr<_T>& refSafePtr)
	{
		if (m_pTarget)
			m_pTarget->RemoveSafePtr(reinterpret_cast<CSafePtr<void>*>(this));
		m_pTarget = refSafePtr.m_pTarget;
		if (m_pTarget)
			m_pTarget->AddSafePtr(reinterpret_cast<CSafePtr<void>*>(this));
		return *this;
	}
	_T* operator -> () const
	{
		if (m_pTarget)
			return m_pTarget;
		else
			throw CString(_T(__FUNCTION__) _T("; Pointer is NULL"));
	}
	operator _T* () const
	{
		return m_pTarget;
	}
private:
	friend class CSafeObject;
	_T*	m_pTarget;
};

/**
 *  @brief	Keep track of referencing CSafePtr's
 */
class CSafeObject
{
public:
	~CSafeObject()
	{
		POSITION pos;
		while ((pos = m_SafePtrList.GetHeadPosition()) != 0)
		{
			CSafePtr<void>* pCurSafePtr = m_SafePtrList.GetAt(pos);
			pCurSafePtr->m_pTarget = NULL;
			m_SafePtrList.RemoveAt(pos);
		}
	}
private:
	friend class CSafePtr;
	void AddSafePtr(CSafePtr<void>* pSafePtr)
	{
		m_SafePtrList.AddTail(pSafePtr);
	}
	void RemoveSafePtr(CSafePtr<void>* pSafePtr)
	{
		POSITION pos;
		for (pos = m_SafePtrList.GetHeadPosition(); pos != 0;)
		{
			CSafePtr<void>* pCurSafePtr = m_SafePtrList.GetAt(pos);
			if (pCurSafePtr == pSafePtr)
			{
				m_SafePtrList.RemoveAt(pos);
				return;
			}
			m_SafePtrList.GetNext(pos);
		}
	}
	CList<CSafePtr<void>*>	m_SafePtrList;
};
